Attribute VB_Name = "WriteStuff"
Option Explicit
Option Base 0
'this module contains routines for the output to a S1M file
Public Cancelled As Boolean
Public UniqueSample(2048)  As Boolean      'unique sample: true when unique
Public UniqueSampleNdx(2048) As Integer
Public nUniqueSamples As Integer       'number of unique samples


Sub WriteS1MInstrument(InstrNr As Integer)
'Write output file for One instrument
'this is the instrument:
    Dim s
    Dim i
    'Call SetProgressBar
    
  
    Set InstrumentNode = fMainForm.TreeView1.Nodes("nodeI" + CStr(InstrNr))
    If InstrumentNode.Children = 0 Then
        s = MsgBox("No samples found", vbCritical, "Error")
        Cancelled = True
        Exit Sub
    End If
    Cancelled = False
    Call OpenFiles
    If Cancelled Then
        Exit Sub
    End If
    Screen.MousePointer = vbHourglass
    nUniqueSamples = 0
    For i = 1 To 2048
        UniqueSample(i) = False
    Next i
    Call CountUniqueSamples(InstrNr)
    Call WriteFileHeader
    Call WriteSize(InstrNr)
    Call WritePointers
    Call WriteWaveHeaders(InstrNr)
    Call WriteSampleHeaders(InstrNr)
    Call WriteRest(InstrNr)
    Call WriteSample(InstrNr)
    Screen.MousePointer = vbDefault
End Sub
Sub OpenFiles()
'open file for input and output
    

    On Error Resume Next   'handel runtime error (=cancel) myself
    's = Mid(InstrumentNode.Text + "        ", 1, 8)
    oFile = shPath + InstrumentNode.Text + ".s1m"
    With fMainForm.dlgCommonDialog
        .DialogTitle = "Save As EX5/7 File"
        .CancelError = True
        .Filter = "S1M Files (*.s1m)|*.s1m"
        .FileName = oFile
        .Flags = cdlOFNOverwritePrompt 'prompt if file exist
        .ShowSave
    End With
        If Err = 20477 Then  'illegal filename
            With fMainForm.dlgCommonDialog
                .DialogTitle = "Save As EX5/7 File"
                .CancelError = True
                .Filter = "S1M Files (*.s1m)|*.s1m"
                .FileName = ""
                .Flags = cdlOFNOverwritePrompt 'prompt if file exist
                .ShowSave
            End With
        End If
        If Err = 32755 Then 'cancel pressed: exit
            Cancelled = True
            Exit Sub
        End If
        If Len(fMainForm.dlgCommonDialog.FileName) = 0 Then 'filename deleted: exit
            Cancelled = True
            Exit Sub
        End If
        oFile = fMainForm.dlgCommonDialog.FileName
    
    Open sFile For Binary Access Read As #1
    Open oFile For Binary Access Write As #2
    fMainForm.sbStatusBar.Panels(1) = "Saving  " + oFile + "..."
    fMainForm.Refresh
End Sub
Sub WriteFileHeader()
    Dim s As String
    s = "SY1200SPA  V1.00"
    Put #2, , s
    s = String(111, &H0)
    s = s + String(1, &H40)
    Put #2, , s
End Sub

Sub WriteSize(InstrNr As Integer)
'write size of waveheaderdata
'caclulate size as follows: 12 bytes per waveheader + 18 bytes per layer (sample)
'e.g. if there is 1 instrument with 4 saples: size is 12 + 4 * 18
'
    Dim n As Long: Dim i As Integer
    Dim size As Long
    Dim s As String
    n = InstrParams(InstrNr).LayerCount       'fout: inclusief global zones
    For i = 1 To 128
        If InstrParams(InstrNr).globalzone(i) Then n = n - 1 'correct for nr of global zones
    Next i
    size = 12 + n * 18
    n = nUniqueSamples                   'number of unique samples
    s = MakeString(size, 4) + MakeString(n, 2)
    Put #2, , s
End Sub
Sub WritePointers()
'write offset of waveheader data
'fixed for 1 instrument: always &H0
Dim s As String: Dim i
    s = String(4, &H0)
    Put #2, , s
    For i = 1 To 1023         'rest
        s = String(4, &HFF)
        Put #2, , s
    Next i
End Sub
Sub WriteWaveHeaders(InstrNr As Integer)
'write wave headers
'fixed for 1 wave
    Dim s As String
    Dim j As Integer
    Dim k As Long
    Dim l As Long
    Dim tel As Long
    Dim nbytes As Long
    Dim teller As Long
    Dim i As Integer
    Dim h As Long
    
    tel = InstrParams(InstrNr).LayerCount
    teller = tel
    For i = 1 To 128  'correct for nr of global zones
        If InstrParams(InstrNr).globalzone(i) Then teller = teller - 1
    Next i
    
'waveheader part
        s = Mid(InstrumentNode.Text + "        ", 1, 8) 'name, 8 bytes, fill with blanks
        s = s + MakeString(teller, 2)   '# layers
        s = s + String(2, &H0)                          'unknown bytes
        
'layer part
        teller = 0                      'counts really written layers (excl. global)
        For j = 1 To tel                'for all samples in wave
            If InstrParams(InstrNr).globalzone(j) Then GoTo skippy   ' skip global zone
            'k = InstrParams(InstrNr).SampleNdx(j) - 1   'base 0
            teller = teller + 1
            k = UniqueSampleNdx(InstrParams(InstrNr).SampleNdx(j)) - 1
            l = k Or &H4000             'k = samplenr, set bit 14 to 1(=RAM sample)
            s = s + MakeString(l, 2)
            s = s + String(1, &H0)      'output level, fixed on 127
            s = s + String(1, &H0)      'unknown, fixed on &H0
            s = s + String(1, InstrParams(InstrNr).LowKey(j))
            s = s + String(1, InstrParams(InstrNr).HighKey(j))
            s = s + String(1, InstrParams(InstrNr).VelLow(j))
            s = s + String(1, InstrParams(InstrNr).VelHigh(j))
            'het lijkt erop of Frequency mode = fixed wordt veroorzaakt door een pan
            'van 128   (bit 8 set)
            h = Int(InstrParams(InstrNr).pan(j) / 500 * 64) + 64 'Pan
            If h > 127 Then h = 127
            s = s + String(1, h)
            h = Int(InstrParams(InstrNr).coarseTune(j) / 100 * 64) + 64
            If h > 127 Then h = 127
            s = s + String(1, h)
            h = Int(InstrParams(InstrNr).fineTune(j) / 100 * 64) + 64
            If h > 127 Then h = 127
            s = s + String(1, h)
            If teller > 1 Then
                s = s + String(1, &H0)      'SamplePlay, already set in layer 1
            Else
                s = s + String(1, &H40)     'SamplePlay, default
            End If
            s = s + String(2, &H0)      'unknown
            s = s + String(2, &H0)      'StartOffset
            s = s + String(2, &H0)      'Loop length
skippy:
        Next j
        Put #2, , s
        nbytes = 12 + teller * 18
        nbytes = 30720 - nbytes         'fixed for 1 wave; should change for more
        s = String(nbytes, &HFF)
        Put #2, , s

End Sub
Sub WriteSampleHeaders(InstrNr As Integer)
'write sample headers
'NB a sample can be more than once in the waveheaders, but only once in the
'sample section!
'Value of start, startloop, end and endloop are cumulative in Soundfont file
'therefore calculate real values by subtracting the startpoint of actual sample
'(values are relative to startpoint)

Dim i As Integer
Dim s As String
Dim tel As Integer
Dim h As Long
Dim nsamples
Dim SStart As Long
Dim SStartLoop As Long
Dim SEndLoop As Long
Dim SEnd As Long
Dim CumCorr As Long                 'cummulative correction
Dim EX5Start(128)
Dim EX5End(128)

    tel = InstrParams(InstrNr).LayerCount
    nsamples = 0
    CumCorr = 0
    For i = 1 To tel                    'process all samples
        If Not UniqueSample(i) Then GoTo skippy 'if sample already processed: skip
        If InstrParams(InstrNr).globalzone(i) Then GoTo skippy  'skip global zone
        nsamples = nsamples + 1
        SStart = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwStart
        SStartLoop = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwStartLoop
        SEndLoop = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwEndLoop
        SEnd = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwEnd
        EX5Start(nsamples) = SStart
        EX5End(nsamples) = SEnd
        'SStart = sfSample(nsamples).dwStart
        'SStartLoop = sfSample(nsamples).dwStartLoop
        'SEndLoop = sfSample(nsamples).dwEndLoop
        'SEnd = sfSample(nsamples).dwEnd
        s = Mid(sfSample(InstrParams(InstrNr).SampleNdx(i)).SampleName + _
            "        ", 1, 8)
        s = s + String(2, &H0)              'wave type, fixed for 16-bit sample
        h = SStartLoop - SStart
        If i > 1 Then
            h = h + EX5End(nsamples - 1)
        End If
        If i > 2 Then
            CumCorr = CumCorr + EX5Start(nsamples - 1) - _
                EX5End(nsamples - 2) - 1
            h = h - CumCorr
        End If
        h = Int(h / 2) Or &HFC0000                  'set bit 18-23
        s = s + MakeString(h, 3)
        Select Case InstrParams(InstrNr).samplemode(i)
            Case 1, 2
                s = s + String(1, &H0)              'FwdLp, frachtion = 0
            'Case 3
                'FwdLpEx  TBD
            Case Else
                s = s + String(1, &H40)             'FwdNoLp, fraction = 0
        End Select
        'If SEnd = SEndLoop Then
        '    s = s + String(1, &H40)             'FwdNoLp, fraction = 0
        'Else
        '    s = s + String(1, &H0)              'FwdLp, frachtion = 0
        'End If
        h = SStartLoop - SStart
        s = s + MakeString(h, 3)
        s = s + String(1, &H0)              'fixed on FwdLp, LoopTune = 0
        h = SEndLoop - SStartLoop
        s = s + MakeString(h, 3)
        If InstrParams(InstrNr).OverridingRootKey(i) = -1 Then
            h = Asc(sfSample(InstrParams(InstrNr).SampleNdx(i)).byOriginalKey)
        Else
            h = InstrParams(InstrNr).OverridingRootKey(i)
        End If
        s = s + MakeString(h, 1)
        h = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwStart - SStart
        s = s + MakeString(h, 3)
        h = SEnd - SEndLoop
        s = s + MakeString(h, 3)      'number of unplayed samples; not in Soundfont?
        s = s + MakeString(nsamples - 1, 2)      'samplenumber, base0
        Put #2, , s
skippy:
    Next i
    
    'fill other sample headers
    nsamples = nsamples + 1                       'from number of samples + 1
    For i = nsamples To 1024
        s = "NewSampl"                  'name
        's = s + MakeString(1, 2)        '# layers?
        s = s + MakeString(&H80, 10)     '6 byte number, fixed &H40
        's = s + MakeString(&H69FC, 3)   'sample length in words?
        s = s + MakeString(&H3C, 4)
        s = s + String(6, &H0)
        s = s + String(2, &HFF)
        Put #2, , s
    Next i
End Sub
Sub WriteRest(InstrNr As Integer)
'write sample seizes in bytes
'NB a sample can be more than once in the waveheaders, but only once in the
'sample section!

Dim i As Integer
Dim s As String
Dim tel As Integer
Dim h As Long
Dim nsamples

    tel = InstrParams(InstrNr).LayerCount
    nsamples = 0
    For i = 1 To tel                    'process all samples
        If Not UniqueSample(i) Then GoTo skippy 'if sample already processed: skip
        If InstrParams(InstrNr).globalzone(i) Then GoTo skippy  'skip global zone
        nsamples = nsamples + 1
        'h = Int(sfSample(InstrParams(InstrNr).SampleNdx(i)).dwEnd - _
                sfSample(InstrParams(InstrNr).SampleNdx(i)).dwStart)
        
        h = Int(sfSample(InstrParams(InstrNr).SampleNdx(i)).dwEnd - _
                sfSample(InstrParams(InstrNr).SampleNdx(i)).dwStart)
        h = h * 2                       'translate in bytes
        s = MakeString(h, 4)
        Put #2, , s
skippy:
    Next i
    
    'fill rest
    nsamples = nsamples + 1                       'from number of samples + 1
    For i = nsamples To 1024
        s = String(4, &H0)
        Put #2, , s
    Next i
End Sub

Sub WriteSample(InstrNr As Integer)
' write samples to file
'NB a sample can be more than once in the waveheaders, but only once in the
'sample section!

    Dim i As Integer
    Dim j As Integer
    Dim s As String
    Dim tel As Integer
    Dim h As Long
    Dim nsamples
    Dim length As Long
    Dim TotLength As Long
    Dim TotCounter As Long
    Dim StartS As Long
    Dim EndS As Long
    Dim counter
    Dim b1 As String: Dim b2 As String

    
    tel = InstrParams(InstrNr).LayerCount
    nsamples = 0
    TotLength = 0: TotCounter = 0
    For i = 1 To tel                    'calculate total length
        If Not UniqueSample(i) Then GoTo skippy1 'if sample already processed: skip
        nsamples = nsamples + 1
        StartS = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwStart * 2 _
            + strtSampleData
        EndS = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwEnd * 2 + strtSampleData
        TotLength = TotLength + (EndS - StartS)
skippy1:
    Next i

    nsamples = 0
    For i = 1 To tel                    'process all samples
        If Not UniqueSample(i) Then GoTo skippy 'if sample already processed: skip
        counter = 1
        nsamples = nsamples + 1
        StartS = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwStart * 2 _
            + strtSampleData
        EndS = sfSample(InstrParams(InstrNr).SampleNdx(i)).dwEnd * 2 + strtSampleData
        length = EndS - StartS
        
        With fMainForm.ProgressBar1
            .Min = 0
            .Max = TotLength + 1
            .Visible = True
        End With
        
        Seek #1, StartS                   'position input file
        Do While counter < length
            b1 = Input(1, #1)       'read little endean
            b2 = Input(1, #1)
            Put #2, , b2
            Put #2, , b1            'write big endean
            counter = counter + 2
            TotCounter = TotCounter + 2
            fMainForm.ProgressBar1.value = TotCounter
        Loop
        'MsgBox ("sample size: " + Str(counter - 1))
skippy:
    Next i
        Close #1
        Close #2
    fMainForm.ProgressBar1.Visible = False
    fMainForm.sbStatusBar.Panels(1) = "File  " + oFile + "  was saved"
End Sub


Function MakeString(var As Long, length As Integer) As String
    'puts var in a string of length bytes, max 128 bytes
    'divided over necessary number of bytes
    Dim Number As Long: Dim Fraction As Long
    Dim s As String: Dim n: Dim j
    Dim b(128)
    Number = var: n = 1
    'divide over n bytes
    Do While Number > 0
        Fraction = Int(Number / &H100)
        b(n) = Number - Fraction * &H100
        Number = Fraction
        n = n + 1
    Loop
    n = n - 1                           'n = number of bytes found > 0
    'compose ouputstring
    s = ""
    For j = 1 To length - n             'fill with &H0 from left
        s = s + String(1, &H0)
    Next j
    
    For j = n To 1 Step -1              'count backwards, lowest bite in b(1) etc
        s = s + String(1, b(j))
    Next j
    MakeString = s
End Function

Sub CountUniqueSamples(InstrNr As Integer)
'count number of unique samples
'a sample can be referenced in a wave more than once, but needs to be written
'only once
'corrected for global zone (26/7/03)

Dim tel
Dim i: Dim j
Dim counter As Integer
tel = InstrParams(InstrNr).LayerCount      'tel = # layers incl. duplicates
    For i = 1 To tel                    'process "all" samples
        For j = 1 To i - 1                'check if sample already processed
            If InstrParams(InstrNr).SampleNdx(i) = InstrParams(InstrNr).SampleNdx(j) _
                Then GoTo skippy              'skip sample
        Next j
        If InstrParams(InstrNr).globalzone(i) Then GoTo skippy 'skip global zone
        UniqueSample(i) = True                    'unique sample found
        counter = counter + 1               'number of unique samples found
        UniqueSampleNdx(InstrParams(InstrNr).SampleNdx(i)) = counter
skippy:
    Next i
    nUniqueSamples = counter
End Sub
